"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.NavigationAdapter = exports.EntityNotFound = void 0;
var ts_odatajs_1 = require("ts-odatajs");
var models_1 = require("../models");
exports.EntityNotFound = 'Could not find entity with type name';
var PartnerSuffix = 'Partner';
var NavigationAdapter = (function () {
    function NavigationAdapter() {
        this.associations = [];
    }
    NavigationAdapter.prototype.adapt = function (metadata) {
        this.metadata = metadata;
        this.entityContainer = ts_odatajs_1.oData.utils.lookupDefaultEntityContainer(this.metadata.schema);
        ts_odatajs_1.oData.utils.forEachSchema(this.metadata.schema, this.adaptSchema.bind(this));
    };
    NavigationAdapter.prototype.adaptSchema = function (schema) {
        var _this = this;
        this.associations = [];
        var entityTypes = schema.entityType || [];
        entityTypes.forEach(function (e) { return _this.adaptEntityType(schema, e); });
        this.setAssociations(schema);
    };
    NavigationAdapter.prototype.adaptEntityType = function (schema, entityType) {
        var _this = this;
        (entityType.navigationProperty || []).forEach(function (n) { return _this.adaptNavigationProperty(schema, entityType.name, n); });
    };
    NavigationAdapter.prototype.adaptNavigationProperty = function (schema, entityTypeName, navigationProperty) {
        var namespace = schema.namespace;
        var navPropType = this.getUnderlyingEntityType(navigationProperty);
        var endpoint = new models_1.AssociationEndpoint({
            containingEntityType: "".concat(namespace, ".").concat(entityTypeName),
            partnerEntityType: navPropType,
            navigationProperty: navigationProperty
        });
        this.trySetReferentialConstraint(endpoint);
        var partnerNavProp = this.tryGetPartnerNavigationProperty(endpoint);
        if (!this.validateRelationship(navigationProperty, partnerNavProp)) {
            partnerNavProp = null;
        }
        var partnerEndpoint = new models_1.AssociationEndpoint({
            containingEntityType: navPropType,
            partnerEntityType: "".concat(namespace, ".").concat(entityTypeName),
            navigationProperty: partnerNavProp,
            propertyName: !!partnerNavProp ? null : "".concat(navigationProperty.name).concat(PartnerSuffix)
        });
        this.trySetReferentialConstraint(partnerEndpoint);
        this.setAssociationSet(namespace, endpoint, partnerEndpoint);
    };
    NavigationAdapter.prototype.getUnderlyingEntityType = function (navigationProperty) {
        var result = ts_odatajs_1.oData.utils.getCollectionType(navigationProperty.type) || navigationProperty.type;
        return result;
    };
    NavigationAdapter.prototype.trySetReferentialConstraint = function (endpoint) {
        if (endpoint.isMapped && endpoint.navigationProperty.referentialConstraint) {
            endpoint.referentialConstraint = endpoint.navigationProperty.referentialConstraint;
            return;
        }
        if (endpoint.isCollection || !NavigationAdapter.inferConstraints) {
            return;
        }
        var entityType = ts_odatajs_1.oData.utils.lookupEntityType(endpoint.containingEntityType, this.metadata.schema);
        var partnerEntityType = ts_odatajs_1.oData.utils.lookupEntityType(endpoint.partnerEntityType, this.metadata.schema);
        if (partnerEntityType === null) {
            throw new Error("".concat(exports.EntityNotFound, " ").concat(endpoint.partnerEntityType));
        }
        var entityName = endpoint.containingEntityShortName;
        var partnerEntityName = endpoint.partnerEntityShortName;
        var referentialConstraint = [];
        var entityKeys = this.getKeys(entityType);
        var entityProperties = this.getProperties(entityType);
        var partnerKeys = this.getKeys(partnerEntityType);
        var partnerProperties = this.getProperties(partnerEntityType);
        partnerKeys.forEach(function (r) {
            var partnerKeyProp = partnerProperties.find(function (p) { return p.name === r.name; });
            var keySuffix = r.name.replace(partnerEntityName, '');
            var possibleFKs;
            if (r.name.toLowerCase() === "".concat(entityName, "Id").toLowerCase()) {
                possibleFKs = entityKeys.map(function (p) { return p.name.toLowerCase(); });
            }
            else {
                possibleFKs = NavigationAdapter.foreignKeyConventions
                    .map(function (func) { return func(endpoint, keySuffix); })
                    .filter(function (p) { return !referentialConstraint.find(function (rc) { return rc.property.toLowerCase() === p; }); });
            }
            var propsMatchingType = entityProperties.filter(function (p) { return p.type === partnerKeyProp.type; });
            var fkProp = possibleFKs
                .map(function (fk) { return propsMatchingType.find(function (p) { return p.name.toLowerCase() === fk; }); })
                .find(function (p) { return !!p; });
            if (fkProp) {
                referentialConstraint.push({ property: fkProp.name, referencedProperty: r.name });
            }
        });
        endpoint.referentialConstraint = referentialConstraint;
    };
    NavigationAdapter.prototype.getAllTypes = function (entityType) {
        var result = [entityType];
        var currentType = entityType;
        while (!!currentType.baseType) {
            var baseType = currentType.baseType;
            currentType = ts_odatajs_1.oData.utils.lookupEntityType(currentType.baseType, this.metadata.schema);
            if (currentType === null) {
                throw new Error("".concat(exports.EntityNotFound, " ").concat(baseType));
            }
            result.push(currentType);
        }
        return result;
    };
    NavigationAdapter.prototype.getKeys = function (entityType) {
        var types = this.getAllTypes(entityType);
        var result = types
            .reduce(function (a, b) { return a.concat((b.key && b.key.propertyRef) || []); }, []);
        return result;
    };
    NavigationAdapter.prototype.getNavigationProperties = function (entityType) {
        var types = this.getAllTypes(entityType);
        var result = types
            .reduce(function (a, b) { return a.concat(b.navigationProperty || []); }, []);
        return result;
    };
    NavigationAdapter.prototype.getProperties = function (entityType) {
        var types = this.getAllTypes(entityType);
        var result = types
            .reduce(function (a, b) { return a.concat(b.property || []); }, []);
        return result;
    };
    NavigationAdapter.prototype.tryGetPartnerNavigationProperty = function (endpoint) {
        var _this = this;
        var navProp = endpoint.navigationProperty;
        var navPropType = this.getUnderlyingEntityType(navProp);
        var partnerEntityType = ts_odatajs_1.oData.utils.lookupEntityType(navPropType, this.metadata.schema);
        if (partnerEntityType === null) {
            throw new Error("".concat(exports.EntityNotFound, " ").concat(navPropType));
        }
        var partnerNameCandidates = [];
        if (navProp.partner) {
            partnerNameCandidates.push(navProp.partner);
        }
        else if (NavigationAdapter.inferPartner) {
            var entitySetName = this.getEntitySetNameByEntityType(endpoint.containingEntityType);
            partnerNameCandidates.push(endpoint.containingEntityShortName);
            partnerNameCandidates.push(entitySetName);
        }
        var partnerNavProperties = this.getNavigationProperties(partnerEntityType);
        var partnerNavProp = partnerNavProperties
            .find(function (n) { return n !== navProp
            && partnerNameCandidates.includes(n.name)
            && (!n.partner || n.partner === navProp.name); });
        if (!!partnerNavProp || !NavigationAdapter.inferPartner) {
            return partnerNavProp;
        }
        partnerNavProp = partnerNavProperties
            .find(function (n) { return n !== navProp
            && _this.navigationPropertyTypeMatches(n, endpoint.containingEntityType)
            && (!n.partner || n.partner === navProp.name); });
        return partnerNavProp;
    };
    NavigationAdapter.prototype.validateRelationship = function () {
        var navigationProperties = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            navigationProperties[_i] = arguments[_i];
        }
        var result = NavigationAdapter.allowManyToMany
            || !navigationProperties.every(function (n) { return !!n && ts_odatajs_1.oData.utils.isCollectionType(n.type); });
        return result;
    };
    NavigationAdapter.prototype.getEntitySetNameByEntityType = function (entityType) {
        var set = this.entityContainer.entitySet.find(function (s) { return s.entityType === entityType; });
        return set && set.name;
    };
    NavigationAdapter.prototype.navigationPropertyTypeMatches = function (navigationProperty, type) {
        var navPropType = this.getUnderlyingEntityType(navigationProperty);
        var result = navPropType === type;
        return result;
    };
    NavigationAdapter.prototype.setAssociationSet = function (namespace, endpoint, partnerEndpoint) {
        var assoc = this.associations.find(function (a) { return (a.containsProperty(partnerEndpoint.navigationProperty)
            || a.containsProperty(endpoint.navigationProperty)); });
        var mappedEndpoints = [endpoint, partnerEndpoint].filter(function (e) { return assoc && assoc.containsProperty(e.navigationProperty); });
        if (mappedEndpoints.length === 2) {
            return;
        }
        var newAssoc = new models_1.AssociationSet(namespace, endpoint, partnerEndpoint);
        if (assoc && !assoc.fullyMapped && newAssoc.fullyMapped) {
            assoc.endpoints = newAssoc.endpoints;
        }
        else if (!assoc || newAssoc.fullyMapped) {
            this.associations.push(newAssoc);
        }
    };
    NavigationAdapter.prototype.setAssociations = function (schema) {
        var _this = this;
        var associations = this.associations.map(function (a) { return _this.processAssociation(a); });
        schema.association = associations;
        this.entityContainer.associationSet = associations;
    };
    NavigationAdapter.prototype.getAssociationEndpoint = function (endpoint) {
        var result = {
            entitySet: this.getEntitySetNameByEntityType(endpoint.partnerEntityType),
            multiplicity: endpoint.multiplicity,
            role: endpoint.role,
            type: endpoint.partnerEntityType
        };
        return result;
    };
    NavigationAdapter.prototype.processAssociation = function (association) {
        var _this = this;
        var result = {
            association: association.associationName,
            name: association.name,
            end: association.endpoints.map(function (e) { return _this.getAssociationEndpoint(e); }),
            referentialConstraint: this.getReferentialConstraint(association)
        };
        association.endpoints.forEach(function (e) {
            if (!e.isMapped) {
                return;
            }
            var partnerEndpoint = association.getPartnerEndpoint(e);
            var navProp = e.navigationProperty;
            if (!!navProp.relationship) {
                return;
            }
            navProp.relationship = association.associationName;
            navProp.fromRole = partnerEndpoint.role;
            navProp.toRole = e.role;
        });
        return result;
    };
    NavigationAdapter.prototype.getReferentialConstraint = function (association) {
        var constraintKeys = {};
        var result = {
            dependent: {
                propertyRef: null,
                role: null
            },
            principal: {
                propertyRef: null,
                role: null
            }
        };
        var principalEndpoint = association.endpoints[0].isCollection ? association.endpoints[1] : association.endpoints[0];
        var dependentEndpoint = association.getPartnerEndpoint(principalEndpoint);
        var constraint = principalEndpoint.referentialConstraint;
        if (constraint.length === 0) {
            return null;
        }
        result.principal.role = principalEndpoint.role;
        result.dependent.role = dependentEndpoint.role;
        constraintKeys[result.dependent.role] = constraint.map(function (p) { return ({ name: p.property }); });
        constraintKeys[result.principal.role] = constraint.map(function (p) { return ({ name: p.referencedProperty }); });
        Object.keys(result).forEach(function (p) {
            var member = result[p];
            member.propertyRef = constraintKeys[member.role];
        });
        return result;
    };
    NavigationAdapter.allowManyToMany = false;
    NavigationAdapter.inferConstraints = true;
    NavigationAdapter.inferPartner = true;
    NavigationAdapter.foreignKeyConventions = [
        function (endpoint, suffix) { return "".concat(endpoint.propertyName).concat(suffix).toLowerCase(); },
        function (endpoint) { return "".concat(endpoint.propertyName, "Id").toLowerCase(); },
        function (endpoint, suffix) { return "".concat(endpoint.partnerEntityShortName).concat(suffix).toLowerCase(); },
        function (endpoint) { return "".concat(endpoint.partnerEntityShortName, "Id").toLowerCase(); }
    ];
    return NavigationAdapter;
}());
exports.NavigationAdapter = NavigationAdapter;
//# sourceMappingURL=navigation-adapter.js.map