"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.OData4UriBuilder = void 0;
var breeze_client_1 = require("breeze-client");
var breeze_odata4_entity_query_1 = require("./breeze-odata4-entity-query");
var breeze_odata4_predicateVisitor_1 = require("./breeze-odata4-predicateVisitor");
var OData4UriBuilder = (function () {
    function OData4UriBuilder() {
        this.name = OData4UriBuilder.BreezeAdapterName;
    }
    OData4UriBuilder.register = function () {
        breeze_client_1.config.registerAdapter('uriBuilder', OData4UriBuilder);
    };
    OData4UriBuilder.prototype.initialize = function () {
        breeze_odata4_predicateVisitor_1.OData4PredicateVisitor.initialize();
    };
    OData4UriBuilder.prototype.buildUri = function (entityQuery, metadataStore) {
        var queryOptions = this.buildQueryOptions({ expand: [] }, entityQuery, metadataStore);
        var qoText = this.toQueryOptionsString(queryOptions);
        entityQuery.resourceName = this.getResource(entityQuery);
        return entityQuery.resourceName + qoText;
    };
    OData4UriBuilder.prototype.buildQueryOptions = function (queryOptions, entityQuery, metadataStore) {
        var entityType = entityQuery._getFromEntityType(metadataStore, false);
        if (!entityType) {
            entityType = new breeze_client_1.EntityType(metadataStore);
        }
        if (entityQuery.inlineCountEnabled) {
            queryOptions.count = true;
        }
        if (entityQuery.skipCount) {
            queryOptions.skip = entityQuery.skipCount;
        }
        if (entityQuery.takeCount != null) {
            queryOptions.top = entityQuery.takeCount;
        }
        this.addExpandOption(entityType, queryOptions, entityQuery.expandClause);
        this.addSelectOption(entityType, queryOptions, entityQuery.selectClause);
        if (entityQuery.parameters[breeze_odata4_entity_query_1.ExpandParamsKey]) {
            this.addExpandOptionsFromSubqueries(entityType, queryOptions, entityQuery.parameters[breeze_odata4_entity_query_1.ExpandParamsKey]);
            delete entityQuery.parameters[breeze_odata4_entity_query_1.ExpandParamsKey];
        }
        queryOptions.filter = this.toWhereODataFragment(entityType, entityQuery.wherePredicate);
        queryOptions.orderby = this.toOrderByODataFragment(entityType, entityQuery.orderByClause);
        return queryOptions;
    };
    OData4UriBuilder.prototype.getResource = function (entityQuery) {
        var resource = entityQuery.resourceName;
        if (!breeze_client_1.core.isEmpty(entityQuery.parameters)) {
            resource = this.formatString(resource, entityQuery.parameters);
        }
        return resource;
    };
    OData4UriBuilder.prototype.getQueryOptions = function (rootOptions, propertyPath) {
        var path = [].concat(propertyPath);
        var rootProperty = path.shift().nameOnServer;
        if (!path.length) {
            return rootOptions;
        }
        var nextExpand = rootOptions.expand.find(function (e) { return e.name === rootProperty; });
        if (!nextExpand) {
            nextExpand = { name: rootProperty, expand: [] };
            rootOptions.expand.push(nextExpand);
        }
        return this.getQueryOptions(nextExpand, path);
    };
    OData4UriBuilder.prototype.addExpandOption = function (entityType, queryOptions, expandClause) {
        var _this = this;
        if (!expandClause) {
            return;
        }
        expandClause.propertyPaths.forEach(function (pp) {
            var props = entityType.getPropertiesOnPath(pp, false, true);
            var workingOptions = _this.getQueryOptions(queryOptions, props);
            var property = props.pop().nameOnServer;
            if (workingOptions.expand.find(function (e) { return e.name === property; })) {
                return;
            }
            var expandOptions = { name: property, expand: [] };
            workingOptions.expand.push(expandOptions);
        });
    };
    OData4UriBuilder.prototype.addExpandOptionsFromSubqueries = function (entityType, queryOptions, subqueries) {
        var _this = this;
        Object.entries(subqueries).forEach(function (_a) {
            var key = _a[0], value = _a[1];
            var props = entityType.getPropertiesOnPath(key, false, true);
            var workingOptions = _this.getQueryOptions(queryOptions, props);
            var property = props.pop().nameOnServer;
            var currentExpand = workingOptions.expand.find(function (e) { return e.name === property; });
            if (!currentExpand) {
                currentExpand = { name: property, expand: [] };
                workingOptions.expand.push(currentExpand);
            }
            _this.buildQueryOptions(currentExpand, value, entityType.metadataStore);
        });
    };
    OData4UriBuilder.prototype.addSelectOption = function (entityType, queryOptions, selectClause) {
        var _this = this;
        if (!selectClause) {
            return;
        }
        selectClause.validate(entityType);
        selectClause.propertyPaths.forEach(function (pp) {
            var props = entityType.getPropertiesOnPath(pp, false, true);
            var workingOptions = _this.getQueryOptions(queryOptions, props);
            workingOptions.select = workingOptions.select || [];
            var property = props.pop().nameOnServer;
            if (workingOptions.select.find(function (p) { return p === property; })) {
                return;
            }
            workingOptions.select.push(property);
        });
    };
    OData4UriBuilder.prototype.toWhereODataFragment = function (entityType, wherePredicate) {
        if (!wherePredicate) {
            return undefined;
        }
        return wherePredicate.toODataFragment({ entityType: entityType });
    };
    OData4UriBuilder.prototype.toOrderByODataFragment = function (entityType, orderByClause) {
        if (!orderByClause) {
            return undefined;
        }
        orderByClause.validate(entityType);
        var orderBy = orderByClause.items.map(function (item) {
            var propertyPath = entityType.clientPropertyPathToServer(item.propertyPath, '/');
            var direction = item.isDesc ? ' desc' : '';
            return propertyPath + direction;
        });
        return orderBy.toString();
    };
    OData4UriBuilder.prototype.toQueryOptionsString = function (queryOptions) {
        var qoStrings = [];
        for (var _i = 0, _a = Object.getOwnPropertyNames(queryOptions); _i < _a.length; _i++) {
            var qoName = _a[_i];
            var qoValue = queryOptions[qoName];
            if (qoValue !== undefined) {
                if (qoName === 'expand') {
                    qoStrings.push(this.getExpandString(qoValue));
                }
                else if (qoName === 'select') {
                    qoStrings.push(this.getSelectString(qoValue));
                }
                else {
                    qoStrings.push("$".concat(qoName, "=").concat(encodeURIComponent(qoValue)));
                }
            }
        }
        qoStrings = qoStrings.filter(function (s) { return !!s; });
        if (qoStrings.length > 0) {
            return '?' + qoStrings.join('&');
        }
        else {
            return '';
        }
    };
    OData4UriBuilder.prototype.getSelectString = function (select) {
        var key = '$select';
        if (!(select && select.length)) {
            return null;
        }
        return "".concat(key, "=").concat(select);
    };
    OData4UriBuilder.prototype.getFilterString = function (filter) {
        var key = '$filter';
        if (!filter) {
            return null;
        }
        return "".concat(key, "=").concat(encodeURIComponent(filter));
    };
    OData4UriBuilder.prototype.getExpandString = function (options) {
        var _this = this;
        var key = '$expand';
        var expandStrings = options.map(function (option) {
            var subOptions = [
                _this.getFilterString(option.filter),
                _this.getSelectString(option.select),
                _this.getExpandString(option.expand)
            ].filter(function (s) { return !!s; });
            if (!subOptions.length) {
                return option.name;
            }
            return "".concat(option.name, "(").concat(subOptions.join(';'), ")");
        });
        var result = expandStrings.length ? "".concat(key, "=").concat(expandStrings) : null;
        return result;
    };
    OData4UriBuilder.prototype.formatString = function (format, params) {
        var props = Object.keys(params);
        var args = props.map(function (prop) { return ({ key: prop, value: params[prop] }); });
        var result = args.reduce(function (formattedString, arg, index) {
            var key = arg.key;
            var value = arg.value;
            if (value == null) {
                value = '';
            }
            var reg = new RegExp("\\{".concat(key, "\\}"), 'gm');
            var newString = formattedString.replace(reg, value);
            if (newString !== formattedString) {
                delete params[key];
            }
            return newString;
        }, format);
        return result;
    };
    OData4UriBuilder.BreezeAdapterName = 'OData4';
    return OData4UriBuilder;
}());
exports.OData4UriBuilder = OData4UriBuilder;
//# sourceMappingURL=breeze-odata4-uriBuilder.js.map